/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */





#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include "liberty.h"
#include "incremental_board.h"

/* For testing incremental board schemes. 0 means the traditional
 * board code. 1 enables the code in incremental_board.c.
 */
#define INCREMENTAL_SCHEME 1

#if INCREMENTAL_SCHEME==0
static int count(int i, int j, int color, char mx[MAX_BOARD][MAX_BOARD],
		 int maxlib, char mark, int libi[], int libj[], int *size);
static int check_for_capture(int m, int n, int color);
static void chain(int i, int j, char ma[MAX_BOARD][MAX_BOARD],
		  char ml[MAX_BOARD][MAX_BOARD], int *adj, int adji[MAXCHAIN],
		  int adjj[MAXCHAIN], int adjsize[MAXCHAIN], 
		  int adjlib[MAXCHAIN], char mark);
static int updateboard(int i, int j, int color);
#endif

/* ================================================================ */
/*                 Board navigation                                 */
/* ================================================================ */


/* Coordinates for the eight directions, ordered
 * south, west, north, east, southwest, northwest, northeast, southeast.
 */
int deltai[8] = { 1,  0, -1,  0,  1, -1, -1, 1};
int deltaj[8] = { 0, -1,  0,  1, -1, -1,  1, 1};


/* ================================================================ */
/*                 Pushing and popping of boards                    */
/* ================================================================ */

#if INCREMENTAL_SCHEME==0
/* Stack of boards for storing positions while reading. */
static board_t stack[MAXSTACK][MAX_BOARD][MAX_BOARD]; 
#endif

/* Stack of trial moves to get to current position 
 * position and which color made them. Perhaps 
 * this should be one array of a structure 
 */
static int      stacki[MAXSTACK];
static int      stackj[MAXSTACK];
static int      move_color[MAXSTACK];
#if INCREMENTAL_SCHEME==0
static int      ko_stacki[MAXSTACK];
static int      ko_stackj[MAXSTACK];      
#endif

static Hash_data  hashdata_stack[MAXSTACK];

#if INCREMENTAL_SCHEME==0
/* Stack of black and white captured stones */
static int      stackbc[MAXSTACK];     
static int      stackwc[MAXSTACK];
#endif

#if INCREMENTAL_SCHEME==0
/*
 * pushgo pushes the position onto the stack.
 */

static void
pushgo(void)
{
  if (showstack)
    gprintf("        *** STACK before push: %d\n", stackp); 
  memcpy(stack[stackp], p, sizeof(p));
  ko_stacki[stackp] = ko_i;
  ko_stackj[stackp] = ko_j;
  memcpy(&hashdata_stack[stackp], &hashdata, sizeof(hashdata));

  stackbc[stackp]=black_captured;
  stackwc[stackp]=white_captured;

  ++stackp;
}
#endif

#if INCREMENTAL_SCHEME==0
/*
 * popgo pops the position from the stack.
 */

void
popgo()
{
  stackp--;
  if (showstack)
    gprintf("<=    *** STACK  after pop: %d\n", stackp);
   
  memcpy(p, stack[stackp], sizeof(p) );
  ko_i = ko_stacki[stackp];
  ko_j = ko_stackj[stackp];

  memcpy(&hashdata, &(hashdata_stack[stackp]), sizeof(hashdata) );

  black_captured=stackbc[stackp];
  white_captured=stackwc[stackp];
  if (count_variations) 
    sgf_write_line(")\n");
}

#endif
#if INCREMENTAL_SCHEME==1

void
popgo()
{
  stackp--;
  if (showstack)
    gprintf("<=    *** STACK  after pop: %d\n", stackp);
  
  incremental_undo_move();
  
  memcpy(&hashdata, &(hashdata_stack[stackp]), sizeof(hashdata));
  if (count_variations) 
    sgf_write_line(")\n");
}

#endif

/*
 * Do the main work of trymove() and tryko(), i.e. the common parts.
 * The ignore_ko flag tells whether an illegal ko capture may be done.
 * Return 1 if the move was valid, otherwise 0.
 */
#if INCREMENTAL_SCHEME==0
static int 
do_trymove(int i, int j, int color, int ignore_ko)
{
  /* 1. The move must be inside the board or i == -1 (pass) */
  assert (i>=0 && i<board_size && j>=0 && j<board_size);

  /* 2. The location must be empty. */
  if (p[i][j] != EMPTY)
    return 0;

  /* 3. The location must not be the ko point, unless ignore_ko==1. */
  if (!ignore_ko
      && i == ko_i && j == ko_j) {
    if (((i>0) && (p[i-1][j] != color))
	|| ((i==0) && (p[i+1][j] != color))) {
      RTRACE("%m would violate the ko rule\n", i, j);
      return 0;
    }
  }
  
  /* So far, so good.  Now push the move on the move stack. */
  stacki[stackp] = i;
  stackj[stackp] = j;
  move_color[stackp] = color;

  /* Check for stack overflow. */
  if (stackp >= MAXSTACK-2) {
    fprintf(stderr, 
	    "gnugo: Truncating search. This is beyond my reading ability!\n");
    return 0;
  }

  /* Ok, the trivial tests went well. Push the position on the stack, 
     and try to actually make the move. */
  pushgo();

  if (verbose == 4)
    dump_stack();

  if (updateboard(i, j, color) == 0) {

    /* no trivial liberties: we need to run approxlib to
     * make sure we do actually have some liberties !
     */

    if (approxlib(i, j, color, 1, NULL, NULL) == 0) {
      RTRACE("%m would be suicide\n", i, j);

      /* Almost do the work of popgo().  The only difference is that
	 we don't close the variations in the sgf file. */
      stackp--;
      if (showstack)
	gprintf("<=    *** STACK  after pop: %d\n", stackp);
   
      memcpy(p, stack[stackp], sizeof(p));
      ko_i = ko_stacki[stackp];
      ko_j = ko_stackj[stackp];
      memcpy(&hashdata, &(hashdata_stack[stackp]), sizeof(hashdata));
      black_captured = stackbc[stackp];
      white_captured = stackwc[stackp];

      return 0;
    }
  }

  return 1;
}

#endif
#if INCREMENTAL_SCHEME==1

static int 
do_trymove(int i, int j, int color, int ignore_ko)
{
  /* We could call legal() here, but that has to do
   * and then un-do all the stuff we want to re-do here.
   * (pushgo, updateboard, ...)  So we replicate its functionality
   * inline.
   */

  /* 1. The move must be inside the board or i == -1 (pass) */
  assert (i>=0 && i<board_size && j>=0 && j<board_size);

  /* 2. The location must be empty. */
  if (p[i][j] != EMPTY)
    return 0;

  /* 3. The location must not be the ko point, unless ignore_ko==1. */
  if (!ignore_ko
      && i == ko_i && j == ko_j) {
    if (((i>0) && (p[i-1][j] != color))
	|| ((i==0) && (p[i+1][j] != color))) {
      RTRACE("%m would violate the ko rule\n", i, j);
      return 0;
    }
  }

  /* 4. Test for suicide. */
  if (incremental_is_suicide(i, j, color)) {
    RTRACE("%m would be suicide\n", i, j);
    return 0;
  }
  
  /* Check for stack overflow. */
  if (stackp >= MAXSTACK-2) {
    fprintf(stderr, 
	    "gnugo: Truncating search. This is beyond my reading ability!\n");
    return 0;
  }

  /* So far, so good. Now push the move on the move stack. These are
   * needed for dump_stack().
   */
  stacki[stackp] = i;
  stackj[stackp] = j;
  move_color[stackp] = color;

  /*
   * FIXME: Do we have to store hashdata in a stack?
   */
  BEGIN_CHANGE_RECORD();
  PUSH_VALUE(ko_i);
  PUSH_VALUE(ko_j);
  memcpy(&hashdata_stack[stackp], &hashdata, sizeof(hashdata));

  ko_i = -1;
  ko_j = -1;
  hashdata_remove_ko(&hashdata);
  
  PUSH_VALUE(black_captured);
  PUSH_VALUE(white_captured);

  if (showstack)
    gprintf("        *** STACK before push: %d\n", stackp);
  ++stackp;

  if (verbose == 4)
    dump_stack();

  incremental_play_move(i, j, color);

  return 1;
}

#endif

/*
 * trymove pushes the position onto the stack, and makes a move
 * at (i, j) of color. Returns one if the move is legal. The
 * stack pointer is only incremented if the move is legal.
 *
 * So the way to use this is:
 *
 *   if (trymove(i, j, color, [message], k, l)) {
 *      ...
 *      popgo();
 *   }   
 *
 * The message can be written as a comment to an sgf file using 
 * sgfdump().  (k, l) can be -1 if they are not needed but if they are 
 * not the location of (k, l) is included in the comment.
 */

int 
trymove(int i, int j, int color, const char *message, int k, int l)
{
  /* Do the real work elsewhere. */
  if (!do_trymove(i, j, color, 0))
    return 0;
  
  /* Store the move in the sgf file if there is a dump file. */
  if (sgf_dump) {
    if (k == -1) {
      if (color == BLACK)
	sgf_write_line("\n(;B[%c%c]C[%s (%d)\n]", 
		       'a'+j, 'a'+i, message, count_variations);
      if (color == WHITE)
	sgf_write_line("\n(;W[%c%c]C[%s (%d)\n]",
		       'a'+j, 'a'+i, message, count_variations);
    } else {
      char movename[4];
	
      if (l < 8)
	*movename = l+65;
      else
	*movename = l+66;
      sprintf(movename+1, "%d", board_size-k);
	
      if (color == BLACK)
	sgf_write_line("\n(;B[%c%c]C[%s at %s (%d)\n]",
		       'a'+j, 'a'+i, message, movename, count_variations);
      if (color == WHITE)
	sgf_write_line("\n(;W[%c%c]C[%s at %s (%d)\n]",
		       'a'+j, 'a'+i, message, movename, count_variations);
    }
  }

  if (count_variations)
    count_variations++;
  stats.nodes++;

  return 1;
}


/*
 * tryko pushes the position onto the stack, and makes a move
 * at (i, j) of color. The move is allowed even if it is an
 * illegal ko capture. It is to be imagined that (color) has
 * made an intervening ko threat which was answered and now
 * the continuation is to be explored.
 *
 * Return 1 if the move is legal with the above caveat. Returns
 * zero if it is not legal because of suicide.
 */

int 
tryko(int i, int j, int color, const char *message)
{
  /* Do the real work elsewhere. */
  if (!do_trymove(i, j, color, 1))
    return 0;

  if (sgf_dump) {
    if (color == BLACK)
      sgf_write_line("\n(;B[tt]C[tenuki (ko threat)]\n\
;W[tt]C[tenuki (answers ko threat)]\n;B[%c%c]C[%s (%d)\n]", 
		     'a'+j, 'a'+i, message, count_variations);
    else if (color == WHITE)
      sgf_write_line("\n(;W[tt]C[tenuki (ko threat)]\n\
;B[tt]C[tenuki (answers ko threat)]\n;B[%c%c]C[%s (%d)\n]",
		       'a'+j, 'a'+i, message, count_variations);
  }
  
  if (count_variations)
    count_variations++;
  stats.nodes++;

  return 1;
}


/*
 * dump_stack() for use under gdb prints the move stack. 
 */

void
dump_stack(void)
{
  int n;

  for (n=0; n<stackp; n++) {
    gprintf("%o%s:%m ", move_color[n] == BLACK ? "B" : "W",
	    stacki[n], stackj[n]);
  }

#if !TRACE_READ_RESULTS
  if (count_variations)
    gprintf("%o (variation %d)", count_variations-1);
#else
  gprintf("%o (%d)", hashdata.hashval);
#endif

  gprintf("%o\n");
}


/* ================================================================ */
/*                         Board manipulation                       */
/* ================================================================ */


/*
 * do_add_stone and do_remove_stone are defined as macros for 
 * efficiency reasons.
 */

#define do_add_stone(i, j, color) \
  do { \
    p[i][j]=color; \
    hashdata_invert_stone(&hashdata, i, j, color); \
  } while (0)

#define do_remove_stone(i, j) \
  do { \
    hashdata_invert_stone(&hashdata, i, j, p[i][j]); \
    p[i][j] = EMPTY; \
  } while (0)


/*
 * And now the functions, accessible from outside this file.
 */

/* place a stone on the board and update the hashdata. */

void
add_stone(int i, int j, int color)
{
  assert(p[i][j] == EMPTY);
  do_add_stone(i, j, color);
#if INCREMENTAL_SCHEME>0
  incremental_invalidate_strings();
#endif
}


/* remove a stone from the board and update the hashdata. */

void
remove_stone(int i, int j)
{
  assert(p[i][j] != EMPTY);
  do_remove_stone(i, j);
#if INCREMENTAL_SCHEME>0
  incremental_invalidate_strings();
#endif
}


/*
 * Remove the string at (i,j), treating it as captured.
 */

void
remove_string(int i, int j)
{
  int color=p[i][j];

  assert(color != EMPTY);

  do_remove_stone(i, j);
  if (color==BLACK)
    black_captured++;
  else
    white_captured++;
  if ((i>0) && (p[i-1][j]==color))
    remove_string(i-1, j);
  if ((i<board_size-1) && (p[i+1][j]==color))
    remove_string(i+1, j);
  if ((j>0) && (p[i][j-1]==color))
    remove_string(i, j-1);
  if ((j<board_size-1) && (p[i][j+1]==color))
    remove_string(i, j+1);
#if INCREMENTAL_SCHEME>0
  incremental_invalidate_strings();
#endif
}


/* ---------------------------------------------------------------- */

#if INCREMENTAL_SCHEME==0
/*
 * Place a "color" on the board at i,j, and remove
 * any captured stones.  If (i, j) == (boardsize, boardsize),
 * the move is a pass, but that is OK too.
 *
 * Assumes that the only stones which can be captured are
 * those immediately adjacent to the piece just played.
 *
 * Returns non-zero if there are any "trivial" liberties, ie if
 * we capture any stones or there are any adjacent liberties.
 * A return value of 0 does not mean that the placed stone has no
 * liberties, though a return of >0 means it definitely does.
 * (ie return of >0 is sufficient, though not necessary, test
 * for move not being a suicide.) It might be more useful to
 * have a return code of 0 mean that the move is definitely
 * suicide, but this would make the function slower.
 *
 * Globals affected: uses count(), so anything it affects
 */

static int
updateboard(int i, int j, int color)
{
  int other = OTHER_COLOR(color);
  int obvious_liberties=0;
  int captures = 0;
  int tot_captures = 0;

#if CHECK_HASHING
  {
    Hashposition pos;
    unsigned long hash;

    /* Check the hash table to see if we have had this position before. */
    board_to_position(p, ko_i, ko_j, /*color,*/ &pos);
    hash = board_hash(p, ko_i, ko_j/*, color*/);
    assert(hash == hashdata.hashval);
    assert(hashposition_compare(&pos, &(hashdata.hashpos)) == 0);
  }
#endif

  last_move_i=i;
  last_move_j=j;

  ko_i = -1;
  ko_j = -1;
  hashdata_remove_ko(&hashdata);

  /* If the move is a pass, we are done here. */
  if (i >= board_size && j >= board_size)
    return -1;

  assert(i>=0 && i<board_size && j>=0 && j<board_size);
  ASSERT( p[i][j] == EMPTY, i, j);

  if (stackp==0) 
    color_has_played|=color;

  do_add_stone(i, j, color);
  DEBUG(DEBUG_BOARD, "Update board: %m = %d\n", i,j, color);

  /* Check to the top. */
  if (i>0) {
    if (p[i-1][j] == other) {
      tot_captures = check_for_capture(i-1, j, other);
      obvious_liberties = tot_captures;
    } else if (p[i-1][j] == EMPTY)
      ++obvious_liberties;
  }

  /* Check to the bottom. */
  if (i<board_size-1) {
    if (p[i+1][j]==other) {
      captures = check_for_capture(i+1, j, other);
      tot_captures += captures;
      obvious_liberties += captures;
    } else if (p[i+1][j] == EMPTY)
      ++obvious_liberties;
  }

  /* Check to the left. */
  if (j>0) {
    if (p[i][j-1]==other) {
      captures = check_for_capture(i, j-1, other);
      obvious_liberties += captures;
      tot_captures += captures;
    } else if (p[i][j-1]==EMPTY)
      ++obvious_liberties;
  }

  /* Check to the right. */
  if (j<board_size-1) {
    if (p[i][j+1]==other) {
      captures = check_for_capture(i, j+1, other);
      obvious_liberties += captures;
      tot_captures += captures;
    } else if (p[i][j+1]==EMPTY)
      ++obvious_liberties;
  }

  /* Check for possible ko and store it into the ko variables. */
  if (tot_captures == 1) {
    int  friends = 0;
    int  libs = 0;
    int  libertyi = -1;
    int  libertyj = -1;

    if (i > 0) {
      if (p[i-1][j] == color) 
	friends++;
      if (p[i-1][j] == EMPTY) {
	libs++;
	libertyi = i - 1;
	libertyj = j;
      }
    }
    if (i < board_size - 1) {
      if (p[i+1][j] == color) 
	friends++;
      if (p[i+1][j] == EMPTY) {
	libs++;
	libertyi = i + 1;
	libertyj = j;
      }
    }
    if (j > 0) {
      if (p[i][j-1] == color) 
	friends++;
      if (p[i][j-1] == EMPTY) {
	libs++;
	libertyi = i;
	libertyj = j - 1;
      }
    }
    if (j < board_size - 1) {
      if (p[i][j+1] == color) 
	friends++;
      if (p[i][j+1] == EMPTY) {
	libs++;
	libertyi = i;
	libertyj = j + 1;

      }
    }

    /* This test guarantees that the newly put stone has exactly one
     * liberty (the newly caught stone) and that it doesn't connect
     * to any friendly stones.  In the latter case it is called 
     * a snapback, and not a ko, which is ok.  
     */
    if (libs == 1 && friends == 0) {
      ko_i = libertyi;
      ko_j = libertyj;
      hashdata_set_ko(&hashdata, libertyi, libertyj);
    }
  }

#if CHECK_HASHING
  {
    Hashposition pos;
    unsigned long hash;

    /* Check the hash table to see if we have had this position before. */
    board_to_position(p, ko_i, ko_j, /*color,*/ &pos);
    hash = board_hash(p, ko_i, ko_j/*, color*/);
    assert(hashposition_compare(&pos, &(hashdata.hashpos)) == 0);
    assert(hash == hashdata.hashval);
  }
#endif

  return obvious_liberties;
}
#endif

/*
 * legal(i, j, color) determines whether the move (color) at
 * (i, j) is legal.
 */

int 
legal(int i, int j, int color)
{
#if INCREMENTAL_SCHEME==0
  int legit=0;
#endif

  /* 0. A pass move is always legal. */
  if (i == board_size && j == board_size)
    return 1;

  /* 1. The move must be inside the board. */
  assert(i>=0 && i<board_size && j>=0 && j<board_size);

  /* 2. The location must be empty. */
  if (p[i][j]!=EMPTY) 
    return 0;

  /* 3. The location must not be the ko point. */
  if (i == ko_i && j == ko_j)
    if (((i>0) && (p[i-1][j] != color))
	|| ((i==0) && (p[i+1][j] != color))) {
      RTRACE("%m would violate the ko rule\n", i, j);
      return 0;
    }

  /* Check for stack overflow. */
  if (stackp >= MAXSTACK-2) {
    fprintf(stderr, 
	    "gnugo: Truncating search. This is beyond my reading ability!\n");
    return 0;
  }

#if INCREMENTAL_SCHEME==0
  /* Ok, the trivial tests went well. Push the position on the stack, 
     and try to actually make the move. */
  pushgo();
  updateboard(i, j, color);

  /* Only care about 0 or >= 1. */
  if (approxlib(i, j, color, 1, NULL, NULL)==0) {
    RTRACE("%m would be suicide\n", i, j);
    legit = 0;
  } else
    legit = 1;

  stackp--;
  if (showstack)
    gprintf("<=    *** STACK  after pop: %d\n", stackp);
   
  memcpy(p, stack[stackp], sizeof(p) );
  ko_i = ko_stacki[stackp];
  ko_j = ko_stackj[stackp];
  memcpy(&hashdata, &(hashdata_stack[stackp]), sizeof(hashdata) );
  black_captured=stackbc[stackp];
  white_captured=stackwc[stackp];

  return legit;
#endif
#if INCREMENTAL_SCHEME==1
  if (incremental_is_suicide(i, j, color)) {
    RTRACE("%m would be suicide\n", i, j);
    return 0;
  }
  return 1;
#endif  
}

/* Play a move. If you want to test for legality you should first call
 * legal(). This function strictly follows the algorithm: 
 * 1. Place a stone of given color on the board.
 * 2. If there are any adjacent opponent strings without liberties,
 *    remove them and increase the prisoner count. 
 * 3. If the newly placed stone is part of a string without liberties,
 *    remove it and increase the prisoner count.
 *
 * In contrast to a move played by trymove() or tryko(), this move
 * can't be unplayed.
 *
 * Notice: This description isn't correct for the old board
 * implementation but will be the future behavior.
 */
void play_move(int i, int j, int color)
{
  assert(stackp == 0);

#if INCREMENTAL_SCHEME==0

  updateboard(i, j, color);

#endif
#if INCREMENTAL_SCHEME==1

#if CHECK_HASHING
  {
    Hashposition pos;
    unsigned long hash;

    /* Check the hash table to see if we have had this position before. */
    board_to_position(p, ko_i, ko_j, /*color,*/ &pos);
    hash = board_hash(p, ko_i, ko_j/*, color*/);
    assert(hash == hashdata.hashval);
    assert(hashposition_compare(&pos, &(hashdata.hashpos)) == 0);
  }
#endif
  last_move_i = i;
  last_move_j = j;
  
  ko_i = -1;
  ko_j = -1;
  hashdata_remove_ko(&hashdata);

  /* If the move is a pass, we are done here. */
  if (i >= board_size && j >= board_size)
    return;

  assert(i>=0 && i<board_size && j>=0 && j<board_size);
  ASSERT( p[i][j] == EMPTY, i, j);

  if (stackp == 0) 
    color_has_played |= color;

  /* Do play the move. */
  incremental_play_move(i, j, color);

#if CHECK_HASHING
  {
    Hashposition pos;
    unsigned long hash;

    /* Check the hash table to see if we have had this position before. */
    board_to_position(p, ko_i, ko_j, /*color,*/ &pos);
    hash = board_hash(p, ko_i, ko_j/*, color*/);
    assert(hashposition_compare(&pos, &(hashdata.hashpos)) == 0);
    assert(hash == hashdata.hashval);
  }
#endif

  incremental_invalidate_strings();

#endif
}


/* ---------------------------------------------------------------- */


/* save_state() and restore_state() backup and restore the board
 * position without using the stack. They are used in counting
 * under the ascii interface by endgame().
 *
 * FIXME: These are currently unused and maybe should be removed. If
 *        they remain they should store ko position and hashdata as
 *        well so they can be used for general purposes.
 */

static board_t p_backup[MAX_BOARD][MAX_BOARD];
static int black_captured_backup;
static int white_captured_backup;


void
save_state()
{
  black_captured_backup=black_captured;
  white_captured_backup=white_captured;
  memcpy(p_backup, p, sizeof(p));
}


void
restore_state()
{
  black_captured=black_captured_backup;
  white_captured=white_captured_backup;
  memcpy(p, p_backup, sizeof(p));
}


/* ---------------------------------------------------------------- */

#if INCREMENTAL_SCHEME==0
/*
 * If the chain at (m,n) has no liberties, remove it from
 * the board. Return the number of stones captured.
 */

static char check_mx[MAX_BOARD][MAX_BOARD];
static char check_mark = -1;

static void
init_check_for_capture(void)
{
  /* set all pieces as unmarked */
  memset(check_mx, 0, sizeof(check_mx));
  check_mark = 1;
}


static int 
check_for_capture(int m, int n, int color)
{
  int i,j;
  int captured=0; /* number of captured */

  ASSERT(p[m][n] == color, m,n);

  DEBUG(DEBUG_COUNT,"Checking %m for capture\n", m,n);

  check_mark++;
  if (check_mark == 0) {
    /* We have wrapped around, reinitialize. */
    init_check_for_capture();
  }

  if (count(m, n, color, check_mx, 1, check_mark, NULL, NULL, NULL) > 0)
    return 0;

  /* Remove all the stones with mx[i][j] == 0,
   *  (since these are the ones count() touched)
   * - count also marks EMPTY stones, but there's
   * no harm in resetting them to EMPTY
   */

  for (i=0; i<board_size; ++i)
    for (j=0; j<board_size; ++j)
      if (check_mx[i][j] == check_mark) {
	/* FIXME: We should really not remove the already removed stones.
	          This seriously upsets the hash code. /iw */
	do_remove_stone(i, j);
	++captured;
      }
  
  if (color == WHITE)
    white_captured += captured;
  else
    black_captured += captured;

  return captured;
}
#endif

/*
 * Initialize board.
 */

void
init_board()
{
  memset(p, EMPTY, sizeof(p));
  ko_i = -1;
  ko_j = -1;
  hashdata_init(&hashdata);
  last_move_i = -1;
  last_move_j = -1;

#if INCREMENTAL_SCHEME == 1
  incremental_board_init();
#endif
}


/*
 * Initialize the gnugo engine.
 *
 * FIXME: This function doesn't belong here. Move it?
 */

void
init_gnugo(float memory)
{
  /* FIXME: The following lines don't actually initialize the board,
     but the engine. */
  /* FIXME2: Moved them to main() in an attempt to fix the broken -M flag.
     Needs to be cleaned up again. */
#if 0
  nominal_depth = DEPTH;
  nominal_backfill_depth = BACKFILL_DEPTH;
  nominal_fourlib_depth = FOURLIB_DEPTH;
  nominal_ko_depth = KO_DEPTH;
#endif
  
#if HASHING
  /* Initialize hash table.
   *
   * The number 1.4 below is the quotient between the number of nodes
   * and the number of read results.  It was found in a test that this 
   * number varies between 1.15 and 1.4.  Thus we use 1.4.
   */
  {
    float nodes;
    
    nodes = ((float) (memory * 1024 * 1024)
	     / (1.5 * sizeof(Hashnode *)
		+ sizeof(Hashnode)
		+ 1.4 * sizeof(Read_result)));
    /* If we get a zero size hash table, disable hashing completely. */
    if (nodes < 1.0)
      hashflags = HASH_NOTHING;
    movehash = hashtable_new((int) 1.5 * nodes,  /* table size   */
			     (int) nodes,        /* nodes        */
			     (int) 1.4 * nodes); /* read results */
  } 
#endif

  init_board();
}


/*
 * Count (minimum) liberty of color piece at (m, n).
 * m,n may contain either that color, or be empty, in which
 * case we get what the liberties would be if we were to place
 * a piece there.
 *
 * If the caller only cares about a minimum liberty count, it
 * should pass that minimum in, and then we *may* stop counting
 * if/when we reach that number. (We may still return a larger
 * number if we feel like it. (The current implementation doesn't.))
 * It's guaranteed that no more than maxlib liberties are written to
 * libi and libj.
 *
 * Note that countlib(i, j, c, libi, libj) is defined as
 * approxlib(i, j, c, MAXLIBS, libi, libj).
 */

#if INCREMENTAL_SCHEME==0
static char approxlib_ml[MAX_BOARD][MAX_BOARD];
static char approxlib_mark = -1;

static void
init_approxlib(void)
{
  /* Set all pieces as unmarked. */
  memset(approxlib_ml,0,sizeof(approxlib_ml));
  approxlib_mark = 1;
}
#endif  

int
approxlib(int m,      /* row number 0 to board_size-1 */
	  int n,      /* column number 0 to board_size-1 */
	  int color,  /* BLACK or WHITE */
	  int maxlib, /* count might stop when it reaches this */
	  int *libi,  /* arrays to put coordinates of liberties into */
	  int *libj)
{
  int lib;

  ASSERT(m>=0 && m < board_size && n >= 0 && n < board_size, m, n);
  ASSERT(color != EMPTY, m, n);
  
  ASSERT(p[m][n] != OTHER_COLOR(color), m,n);

#if INCREMENTAL_SCHEME == 0
  
  approxlib_mark++;
  if (approxlib_mark == 0) { 
    /* We have wrapped around, reinitialize. */
    init_approxlib();
  }

  /* Count liberty of current piece. */
  lib = count(m, n, color, approxlib_ml, maxlib, approxlib_mark,
	      libi, libj, NULL);

#endif
#if INCREMENTAL_SCHEME == 1
  
  if (libi)
    lib = incremental_findlib(m, n, color, maxlib, libi, libj);
  else
    lib = incremental_countlib(m, n, color, maxlib);
  
#endif
  
  DEBUG(DEBUG_COUNT, "approxlib(%m,%d) returns %d\n", m, n, maxlib, lib);
  return lib;
}


/* Count the number of stones in a string. */
int
countsize(int m, int n)
{
  int size;

  ASSERT(m>=0 && m < board_size && n >= 0 && n < board_size, m, n);
  ASSERT(p[m][n] != EMPTY, m, n);

#if INCREMENTAL_SCHEME == 0
  approxlib_mark++;
  if (approxlib_mark == 0) { 
    /* We have wrapped around, reinitialize. */
    init_approxlib();
  }

  /* Count the size of the current piece. */
  count(m, n, p[m][n], approxlib_ml, MAXLIBS, approxlib_mark,
	NULL, NULL, &size);

#endif
#if INCREMENTAL_SCHEME == 1

  size = incremental_countsize(m, n);
  
#endif
  
  DEBUG(DEBUG_COUNT, "countsize(%m) returns %d\n", m, n, size);
  
  return size;
}

#if INCREMENTAL_SCHEME==0
/*
 * Count (minimum) liberty of color piece at location (i, j)
 * and return value in global variable lib. Return size of
 * connected component in size.
 *
 * No check is made on entry that that square is occupied
 * by that color; It is valid to use this to count the
 * liberties that there would be if color was put into
 * empty square at (i,j).
 *
 * If k<lib, then (libi[k], libj[k]) points to one of the
 * liberties of the string.
 *
 * ENTRY: mark is the current value of "mark"
 *
 *        mx[][] contains "mark" for stones and empty intersections that
 *	  may not be counted. Typically "mark" is a value that at calling
 *	  time is absent from mx[][].
 *
 *        maxlib: Typically, it is not necessary to get the
 *        exact number of liberties: we can stop when it
 *        reaches some threshold (often 1 !).
 *
 *        libi, libj: Arrays to put liberty locations in. These should be
 *        dimensioned to hold up to maxlib locations. If libi is NULL,
 *        the locations are not returned and possibly not computed at all.
 *
 *        size: Variable to put string size into, may be NULL if we don't
 *        care.
 *
 * RETURNS: number of liberties found
 *
 * EXIT:  updates mx[][], putting "mark" at all stones and empty
 *	  intersections visited
 *
 *	  EITHER : return < maxlib, and mx[][] and l[][] fully updated
 *	  OR     : return == maxlib, and mx[][] and l[][] partially updated
 *
 * NOTE: intended as an internal helper for approxlib, but
 *       is exported as a general (though complex) function
 *       for efficiency.
 */

static int count_stacki[MAXCHAIN];
static int count_stackj[MAXCHAIN];

static int
count(int i,     /* row number 0 to board_size-1 */
      int j,     /* column number 0 to board_size-1 */
      int color, /* BLACK or WHITE */
      char mx[MAX_BOARD][MAX_BOARD],  /* workspace: mx[][]=mark means stone already visited. */
      int maxlib, /* Often, we only care if liberties > some minimum (usually 1) */
      char mark, /* Value to mark visited points with. */
      int *libi, /* arrays to put coordinates of liberties into */
      int *libj,
      int *size) /* variable to put string size into */
{
  int count_stackp;
  int lib = 0;
  
  mx[i][j] = mark;
  count_stacki[0] = i;
  count_stackj[0] = j;
  count_stackp = 1;
  if (size)
    *size = 0;
  
  while (count_stackp>0) {
    count_stackp--;
    i = count_stacki[count_stackp];
    j = count_stackj[count_stackp];
    if (size)
      (*size)++;

    /* check West neighbor */
    if (j != 0 && mx[i][j-1] != mark) {
      if (p[i][j-1] == EMPTY) {
	mx[i][j-1] = mark;
	if (libi) {
	  libi[lib] = i;
	  libj[lib] = j-1;
	}
	++lib;
	if (lib == maxlib)
	  return lib;
      }
      else if (p[i][j-1] == color) {
	mx[i][j-1] = mark;
	count_stacki[count_stackp] = i;
	count_stackj[count_stackp] = j-1;
	count_stackp++;
      }
    }
    
    while(1) {
      /* check North neighbor */
      if (i != 0 && mx[i-1][j] != mark) {
	if (p[i-1][j] == EMPTY) {
	  mx[i-1][j] = mark;
	  if (libi) {
	    libi[lib] = i-1;
	    libj[lib] = j;
	  }
	  ++lib;
	  if (lib == maxlib)
	    return lib;
	} 
	else if (p[i-1][j] == color) {
	  mx[i-1][j] = mark;
	  count_stacki[count_stackp] = i-1;
	  count_stackj[count_stackp] = j;
	  count_stackp++;
	}
      }

      /* check South neighbor */
      if (i != board_size-1 && mx[i+1][j] != mark) {
	if (p[i+1][j] == EMPTY) {
	  mx[i+1][j] = mark;
	  if (libi) {
	    libi[lib] = i+1;
	    libj[lib] = j;
	  }
	  ++lib;
	  if (lib == maxlib)
	    return lib;
	}
	else if (p[i+1][j] == color) {
	  mx[i+1][j] = mark;
	  count_stacki[count_stackp] = i+1;
	  count_stackj[count_stackp] = j;
	  count_stackp++;
	}
      }

      /* check East neighbor */
      if (j == board_size-1 || mx[i][j+1] == mark)
	break;
      if (p[i][j+1] == EMPTY) {
	mx[i][j+1] = mark;
	if (libi) {
	  libi[lib] = i;
	  libj[lib] = j+1;
	}
	++lib;
	if (lib == maxlib)
	  return lib;
	break;
      }
      else if (p[i][j+1] == color) {
	mx[i][j+1] = mark;
	if (size)
	  (*size)++;
	j++;
      }
      else
	break;
    }
  }

  return lib;
}
#endif

/* chainlinks returns (in adji, adjj arrays) the chains surrounding
 * the group at i, j. Adapted from count. Marks only one stone on
 * each link. If stackp <= depth, these are sorted by size (largest
 * first).
 *
 * FIXME: The sorting isn't implemented. Should it be?
 */

#if INCREMENTAL_SCHEME==0
static char chainlinks_ma[MAX_BOARD][MAX_BOARD];
static char chainlinks_ml[MAX_BOARD][MAX_BOARD];
static char chainlinks_mark = -1;

static void
init_chainlinks(void)
{
  /* set all pieces as unmarked */
  memset(chainlinks_ma, 0, sizeof(chainlinks_ma));
  memset(chainlinks_ml, 0, sizeof(chainlinks_ml));
  chainlinks_mark = 1;
}
#endif

void 
chainlinks(int m, int n, int *adj, int adji[MAXCHAIN], int adjj[MAXCHAIN],
	   int adjsize[MAXCHAIN], int adjlib[MAXCHAIN])
{
#if INCREMENTAL_SCHEME == 0
  chainlinks_mark++;
  if (chainlinks_mark == 0) { /* We have wrapped around, reinitialize. */
    init_chainlinks();
  }
  
  (*adj)=0;
  
  chain(m, n, chainlinks_ma, chainlinks_ml, adj, adji, adjj, adjsize, adjlib,
	chainlinks_mark);
#endif
#if INCREMENTAL_SCHEME == 1
  incremental_chainlinks(m, n, adj, adji, adjj, adjsize, adjlib);
#endif
}


#if INCREMENTAL_SCHEME==0
/* worker fn for chainlinks.
 * ma and ml store state about where we have been.
 * ma stores which elements of the surrounded chain have been visited.
 * ml (via count()) stores which surrounding stones have been visited.
 *
 * Algorithm :
 *   Mark ma[i][j] as done.
 *   explore the stones surrounding i,j :
 *    - for any which are the same color and not yet visited,
 *      save them in an internal stack for later visit.
 *    - any which are the opposite colour, make a note if not visited,
 *      and use count() to mark them all as visited.
 */

static int seedi[MAXCHAIN];
static int seedj[MAXCHAIN];
static int chain_libi[MAXLIBS];
static int chain_libj[MAXLIBS];

static void 
chain(int i, int j, char ma[MAX_BOARD][MAX_BOARD],
      char ml[MAX_BOARD][MAX_BOARD], int *adj, int adji[MAXCHAIN],
      int adjj[MAXCHAIN], int adjsize[MAXCHAIN], int adjlib[MAXCHAIN],
      char mark)
{
  int k;
  int seedp;
  int color=p[i][j];
  int other=OTHER_COLOR(color);
  int size;

  seedi[0] = i;
  seedj[0] = j;
  seedp = 1;

  while (seedp>0) {
    seedp--;
    i = seedi[seedp];
    j = seedj[seedp];

    /* We may already have visited here. */
    if (ma[i][j] == mark)
      continue;
    ma[i][j] = mark;

    /* check North neighbor */
    if (i != 0) {
      if ((p[i-1][j] == other) && ma[i-1][j] != mark && ml[i-1][j] != mark) {
	adjlib[*adj] = count(i-1, j, other, ml, MAXLIBS, mark,
			     chain_libi, chain_libj, &size);
	for(k=0; k<adjlib[*adj]; k++)
	  ml[chain_libi[k]][chain_libj[k]] = mark-1;
	adji[*adj] = i-1;
	adjj[*adj] = j;
	adjsize[*adj] = size;
	++(*adj);
	ma[i-1][j] = mark;
      } 
      else {
	if ((p[i-1][j] == color) && ma[i-1][j] != mark) {
	  seedi[seedp] = i-1;
	  seedj[seedp] = j;
	  seedp++;
	}
      }
    }
    /* check South neighbor */
    if (i != board_size-1) {
      if ((p[i+1][j] == other) && ma[i+1][j] != mark && ml[i+1][j] != mark) {
	adjlib[*adj] = count(i+1, j, other, ml, MAXLIBS, mark,
			     chain_libi, chain_libj, &size);
	for(k = 0; k < adjlib[*adj]; k++)
	  ml[chain_libi[k]][chain_libj[k]] = mark-1;
	adji[*adj] = i+1;
	adjj[*adj] = j;
	adjsize[*adj] = size;
	++(*adj);
	ma[i+1][j] = mark;
      }
      else {
	if ((p[i+1][j] == color) && ma[i+1][j] != mark) {
	  seedi[seedp] = i+1;
	  seedj[seedp] = j;
	  seedp++;
	}
      }
    }
    /* check West neighbor */
    if (j != 0) {
      if ((p[i][j-1] == other) && ma[i][j-1] != mark && ml[i][j-1] !=mark) {
	adjlib[*adj] = count(i, j-1, other, ml, MAXLIBS, mark,
			     chain_libi, chain_libj, &size);
	for(k = 0; k < adjlib[*adj]; k++)
	  ml[chain_libi[k]][chain_libj[k]] = mark-1;
	adji[*adj] = i;
	adjj[*adj] = j-1;
	adjsize[*adj] = size;
	++(*adj);
	ma[i][j-1] = mark;
      }
      else {
	if ((p[i][j-1] == color) && ma[i][j-1] != mark) {
	  seedi[seedp] = i;
	  seedj[seedp] = j-1;
	  seedp++;
	}
      }
    }
    /* check East neighbor */
    if (j != board_size-1) {
      if ((p[i][j+1] == other) && ma[i][j+1] != mark && ml[i][j+1] != mark) {
	adjlib[*adj] = count(i, j+1, other, ml, MAXLIBS, mark,
			     chain_libi, chain_libj, &size);
	for(k = 0; k < adjlib[*adj]; k++)
	  ml[chain_libi[k]][chain_libj[k]] = mark-1;
	adji[*adj] = i;
	adjj[*adj] = j+1;
	adjsize[*adj] = size;
	++(*adj);
	ma[i][j+1] = mark;
      }
      else {
	if ((p[i][j+1] == color) && ma[i][j+1] != mark) {
	  seedi[seedp] = i;
	  seedj[seedp] = j+1;
	  seedp++;
	}
      }
    }
  }
} 
#endif

/*
 * Find the origin of a worm or a cavity, i.e. the point with smallest
 * i coordinate and in the case of a tie with smallest j coordinate.
 * The idea is to have a canonical reference point for a string.
 */

#if INCREMENTAL_SCHEME==0
static int origin_stacki[MAXCHAIN];
static int origin_stackj[MAXCHAIN];
static char origin_mx[MAX_BOARD][MAX_BOARD];
static char origin_mark = -1;
#endif

void
find_origin(int i, int j, int *origini, int *originj)
{
#if INCREMENTAL_SCHEME == 1
  incremental_find_origin(i, j, origini, originj);
#endif
#if INCREMENTAL_SCHEME == 0
  int origin_stackp;
  int color = p[i][j];

  origin_mark++;
  if (origin_mark == 0) {
    /* We have wrapped around, reinitialize. */
    memset(origin_mx, 0, sizeof(origin_mx));
    origin_mark = 1;
  }
  
  *origini = i;
  *originj = j;
  
  origin_mx[i][j] = origin_mark;
  origin_stacki[0] = i;
  origin_stackj[0] = j;
  origin_stackp = 1;
  
  while (origin_stackp>0) {
    origin_stackp--;
    i = origin_stacki[origin_stackp];
    j = origin_stackj[origin_stackp];

    /* Is this point a possible origin? */
    if (i < *origini || (i == *origini && j < *originj)) {
      *origini = i;
      *originj = j;
    }
    
    /* Check West neighbor. */
    if (j != 0 && origin_mx[i][j-1] != origin_mark) {
      if (p[i][j-1] == color) {
	origin_mx[i][j-1] = origin_mark;
	origin_stacki[origin_stackp] = i;
	origin_stackj[origin_stackp] = j-1;
	origin_stackp++;
      }
    }
    
    while(1) {
      /* Check North neighbor. */
      if (i != 0 && origin_mx[i-1][j] != origin_mark) {
	if (p[i-1][j] == color) {
	  origin_mx[i-1][j] = origin_mark;
	  origin_stacki[origin_stackp] = i-1;
	  origin_stackj[origin_stackp] = j;
	  origin_stackp++;
	}
      }

      /* Check South neighbor. */
      if (i != board_size-1 && origin_mx[i+1][j] != origin_mark) {
	if (p[i+1][j] == color) {
	  origin_mx[i+1][j] = origin_mark;
	  origin_stacki[origin_stackp] = i+1;
	  origin_stackj[origin_stackp] = j;
	  origin_stackp++;
	}
      }

      /* Check East neighbor. */
      if (j == board_size-1 || origin_mx[i][j+1] == origin_mark)
	break;
      if (p[i][j+1] == color) {
	origin_mx[i][j+1] = origin_mark;
	j++;
      }
      else
	break;
    }
  }
#endif
}





/*
 * Return true if the move (i,j) by (color) is a ko capture
 * (whether capture is legal on this move or not).
 */

int
is_ko(int i, int j, int color)
{
#if INCREMENTAL_SCHEME == 1
  return incremental_is_ko(i, j, color);
#endif
#if INCREMENTAL_SCHEME == 0
  int other=OTHER_COLOR(color);
  int captures=0;

  if (i>0) {
    if (p[i-1][j] != other)
      return 0;
    if (approxlib(i-1, j, other, 2, NULL, NULL)==1) {
      if (!singleton(i-1, j))
	return 0;
      captures++;
    }
  }

  if (i<board_size-1) {
    if (p[i+1][j] != other)
      return 0;
    if (approxlib(i+1, j, other, 2, NULL, NULL)==1) {
      if (!singleton(i+1, j))
	return 0;
      captures++;
    }
  }
  
  if (j>0) {
    if (p[i][j-1] != other) 
      return 0;
    if (approxlib(i, j-1, other, 2, NULL, NULL)==1) {
      if (!singleton(i, j-1))
	return 0;
      captures++;
    }
  }
  
  if (j<board_size-1) {
    if (p[i][j+1] != other)
      return 0;
    if (approxlib(i, j+1,other, 2, NULL, NULL)==1) {
      if (!singleton(i, j+1))
	return 0;
      captures++;
    }
  }

  return (captures==1);
#endif
}

/*
 * Returns true if there is an isolated stone (string of size 1) at (i, j).
 */

int
singleton(int i, int j)
{
  if (p[i][j]==EMPTY)
    return 0;
  if ((i>0) && (p[i-1][j]==p[i][j]))
    return 0;
  if ((i<board_size-1) && (p[i+1][j]==p[i][j]))
    return 0;
  if ((j>0) && (p[i][j-1]==p[i][j]))
    return 0;
  if ((j<board_size-1) && (p[i][j+1]==p[i][j]))
    return 0;
  return 1;
}




/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
